------------------------------
--      Are you local?      --
------------------------------

local L = AceLibrary("AceLocale-2.2"):new("BagginsAnywhereBags")
local abacus = AceLibrary("Abacus-2.0")

local _G = getfenv(0)
local type = type
local pairs = pairs
local tonumber = tonumber
local select = select
local playerName = UnitName("player")
local GetMoney = GetMoney

local bagframes = {}

local r

----------------------------
--      Localization      --
----------------------------

L:RegisterTranslations("enUS", function() return {	
	["Anywhere Bags"] = true,
	["View any character's bank and bags from anywhere."] = true,
	["Sort Type"] = true,
	["Sort which way to show items."] = true,
	["Compressed"] = true,
	["Compress items."] = true,
	["Search"] = true,
	["Search for items containing text and output to chat frame."] = true,
	["|cff7fff7fSearching all players for:|r "] = true,
	["|cffff0000No items found.|r"] = true,
	["Bags"] = true,
	["Groups"] = true,
	["Sections"] = true,
	[" Equiped"] = true,
	[" Keyring"] = true,
	[" Mail"] = true,
	["Unknown bagid"] = true,
	["Unknown sectionid"] = true,

	-- bag names
	["Keyring"] = true,
	["Bank"] = true, --Bank name and Bank bags prefix
	["Backpack"] = true,
	["InvBag1"] = true,
	["InvBag2"] = true,
	["InvBag3"] = true,
	["InvBag4"] = true,
	["BankBag1"] = true, --Must start with Bank bags prefix
	["BankBag2"] = true, --Must start with Bank bags prefix
	["BankBag3"] = true, --Must start with Bank bags prefix
	["BankBag4"] = true, --Must start with Bank bags prefix
	["BankBag5"] = true, --Must start with Bank bags prefix
	["BankBag6"] = true, --Must start with Bank bags prefix
	["BankBag7"] = true, --Must start with Bank bags prefix
	["Equiped"] = true,
	["Mail"] = true,
} end)

---------------------------------
--      AceOptions Config      --
---------------------------------

local options = {
	name = L["Anywhere Bags"], type="group",
	desc = L["View any character's bank and bags from anywhere."],
	order = 200,
	args = {
		search = {
			name = L["Search"], type = "text",
			desc = L["Search for items containing text and output to chat frame."],
			get = false,
			set = function(v) BagginsAnywhereBags:Search(v) end,
			usage = "<itemLink, itemName, itemID, or pattern>",
			validate = function(v) return true end,
			order = -40,
		},
		spacer = {
			type = "header",
			order = -30,
		},
		sortType = {
			name = L["Sort Type"], type = "text",
			desc = L["Sort which way to show items."],    
			get = function() return BagginsAnywhereBags.db.profile.sortType end,
			set = function(v) 
				BagginsAnywhereBags.db.profile.sortType = v 
				BagginsAnywhereBags:RebuildOptions()
			end,
			validate = {L["Bags"], L["Groups"], L["Sections"]},
			order = -20,
		},
		compressed = {
			name = L["Compressed"], type = "toggle",
			desc = L["Compress items."],   
			get = function() return BagginsAnywhereBags.db.profile.compressed end,
			set = function(v) 
				BagginsAnywhereBags.db.profile.compressed = v 
				BagginsAnywhereBags:RebuildOptions()
			end,
			order = -10,
		},
	},
}

---------------------------------
--      Addon Declaration      --
---------------------------------
--BAB = BagginsAnywhereBags
BagginsAnywhereBags = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceDB-2.0", "AceConsole-2.0", "AceDebug-2.0", "AceHook-2.1")
BagginsAnywhereBags.revision = tonumber(string.sub("$Revision: 46396 $", 12, -3))

local bagNames = {
	[-2]=L["Keyring"],[-1]=L["Bank"],[0]=L["Backpack"],L["InvBag1"],
	L["InvBag2"],L["InvBag3"],L["InvBag4"],L["BankBag1"],
	L["BankBag2"],L["BankBag3"],L["BankBag4"],L["BankBag5"],
	L["BankBag6"],L["BankBag7"],Equiped=L["Equiped"],Mail=L["Mail"],
}

------------------------------
--      Initialization      --
------------------------------

function BagginsAnywhereBags:OnInitialize()
	self:RegisterDB("BagginsAnywhereBagsDB")
	BagginsAnywhereBags:RegisterDefaults('realm', {
		[playerName] = {
			[-2] = {}, [-1] = {}, [0] = {},
			-- 1 to 11:
			{},{},{},{},{},{},{},{},{},{},{},
			Equiped = {},
			Money = 0,
			Mail = {},
		},
		DBVersion = "1", --TODO: testing code, remove later
	})
	BagginsAnywhereBags:RegisterDefaults('profile', {
		sortType = L["Groups"],
		compressed = true,
		DBVersion = "1", --TODO: testing code, remove later
	})
end

--[[----------------

items stored as:

BagginsAnywhereBags.db.realm[player][bag][slot] = {itemLink, count, texture, quality, Baggins_bagid, Baggins_sectionid}

----------------]]--

function BagginsAnywhereBags:OnEnable(first)
	if not Baggins then return end
	
	if first then
		self:RegisterChatCommand({"/BagginsAnywhereBags", "/BAB"}, {
			name = options.name, type = "group",
			desc = options.desc,
			args = {
				search = options.args.search,
				sortType = options.args.sortType,
				compressed = options.args.compressed,
			},
		})
	end
	
	r = self.db.realm
	local p = self.db.profile
	
	--bank and money need stored whenever they change
	self:RegisterEvent("BANKFRAME_CLOSED", "BankUpdate") -- fired twice
	self:RegisterEvent("PLAYER_MONEY", "MoneyUpdate")
	
	--figure out how to record mail
	self:RegisterEvent("MAIL_CLOSED", "MailUpdate")
	
	--only care about storing bags when logging out (you can always access your own bags)
	if type(_G["Logout"]) == "function" then
		self:Hook("Logout", "OnLogout", true)
	end
	
	if type(_G["Quit"]) == "function" then
		self:Hook("Quit", "OnQuit", true)
	end
	
	r[playerName].Money = GetMoney()
	
	self:RebuildOptions()
	
	Baggins.OnMenuRequest.args.anywherebags = options
end

function BagginsAnywhereBags:OnDisable()
	if not Baggins then return end
	Baggins.OnMenuRequest.args.anywherebags = nil
end

------------------------------
--      Core Functions      --
------------------------------

function BagginsAnywhereBags:ProcessBag(bagid)
	local texture, count, quality, link
	
	r[playerName][bagid] = {}

	for i = 1, GetContainerNumSlots(bagid) do
		texture, count = GetContainerItemInfo(bagid, i);
		if texture then
			link = GetContainerItemLink(bagid, i)
			quality = select(3,GetItemInfo(link))
			r[playerName][bagid][i] = {link, count, texture, quality, self:GetBagginsIDs(bagid, i)}
		else
			r[playerName][bagid][i] = nil
		end
	end
end

function BagginsAnywhereBags:ProcessEquiped()
	local link
	for i = 0, 19 do
		link = GetInventoryItemLink("player", i)
		if link then
			r[playerName].Equiped[i] = {
				link,
				GetInventoryItemCount("player",i),
				GetInventoryItemTexture("player",i),
				GetInventoryItemQuality("player",i),
				playerName..L[" Equiped"],
				playerName..L[" Equiped"],
			}
		else
			r[playerName].Equiped[i] = nil
		end
	end
end

function BagginsAnywhereBags:OnQuit(...)
	if r[playerName] then
		self:FullProcess()
	end
	self.hooks["Quit"](...)
end

function BagginsAnywhereBags:OnLogout(...)
	if r[playerName] then
		self:FullProcess()
	end
	self.hooks["Logout"](...)
end

function BagginsAnywhereBags:FullProcess()
	self:ProcessBag(-2)
	for i = 0,4 do
		self:ProcessBag(i)
	end
	self:ProcessEquiped()
	self:MoneyUpdate()
end

function BagginsAnywhereBags:GetBagginsIDs(bag, slot)
	if bag == -2 then
		return playerName..L[" Keyring"], playerName..L[" Keyring"]
	end
	
	local key = bag..":"..slot
	
	local IsSlotMine = function(mybagid, mysecid, key)
		for bagid, bagframe in ipairs(Baggins.bagframes) do
			for sectionid, section in ipairs(bagframe.sections) do
				if section.slots[key] then
					return mybagid == bagid and mysecid == sectionid
				end
			end
		end
	end
	
	for bagid, bag in pairs(Baggins.db.profile.bags) do
		for sectionid, section in pairs(bag.sections) do
			if IsSlotMine(bagid, sectionid, key) then
				return bag.name, section.name
			end
		end
	end
	
	return L["Unknown bagid"], L["Unknown sectionid"]
end

function BagginsAnywhereBags:Search(v)
	local found = false
	local p = self.db.profile
	local itemName
	local counts = {}
	
	v = string.lower(v)
	
	DEFAULT_CHAT_FRAME:AddMessage(L["|cff7fff7fSearching all players for:|r "]..v)
	for player, bags in pairs(r) do
		if type(bags) == "table" then
			for bag,slots in pairs(bags) do
				if type(slots) == "table" then
					for slot,item in pairs(slots) do
						if type(item) == "table" then
							if string.find(string.lower(item[1]), v) then
								itemName = player..select(3,string.find(item[1], "^|c%x+|H.+|h%[(.+)%]"))
								if type(counts[player]) ~= "table" then
									counts[player] = {}
								end
								if type(counts[player][itemName]) ~= "table" then
									counts[player][itemName] = {"|cff7fff7f["..player.."]:|r "..item[1].."x", item[2]}
								else
									counts[player][itemName][2] = counts[player][itemName][2] + item[2]
								end
								found = true
							end
						end
					end
				end
			end
		end
	end
	if found then	
		for _,v1 in pairs(counts) do
			for _,v in pairs(v1) do
				DEFAULT_CHAT_FRAME:AddMessage(v[1]..v[2])
			end
		end
	else
		DEFAULT_CHAT_FRAME:AddMessage(L["|cffff0000No items found.|r"])
	end
end

-------------------------------
--      Event Functions      --
-------------------------------

do
	local bankAccess = true --only want to access every other one
	function BagginsAnywhereBags:BankUpdate()
		bankAccess = not bankAccess
		if bankAccess then return end
		self:ProcessBag(-1)
		for i = 5,11 do
			self:ProcessBag(i)
		end
		self:RebuildOptions()
	end
end

function BagginsAnywhereBags:MoneyUpdate()
	local m = GetMoney()
	r[playerName].Money = m
end

do
   local mailAccess = true --only want to access every other one
   function BagginsAnywhereBags:MailUpdate()
      mailAccess = not mailAccess
      if mailAccess then return end
      
      r[playerName][L["Mail"]] = {}
      local num = GetInboxNumItems()
      
      local count, link, quality, texture
      for i = 1,num do
         link = GetInboxItemLink(i)
         
         if link then
            count = select(3,GetInboxItem(i))
            quality,_,_,_,_,_,_,texture = select(3,GetItemInfo(link))
            tinsert(r[playerName][L["Mail"]], {
					link,
					count,
					texture,
					quality,
					playerName..L[" Mail"],
					playerName..L[" Mail"],
				})
         end
      end
      self:RebuildOptions()
   end
end

---------------------------------
--      Display Functions      --
---------------------------------

function BagginsAnywhereBags:RebuildOptions()
	local bagName, disabled, opBag, count, itemName, itemLink, bagid, sectionid
	local p = self.db.profile
	
	for player,bags in pairs(r) do
		if type(bags) == "table" then --TODO: testing code, remove check later (Checking because of DBVersion)
			options.args[player] = {
				name = player, type = "group",
				desc = player,
				args = {},
			}
			for bag,slots in pairs(bags) do
				if type(slots) == "table" then
					bagName = bagNames[bag]
					if p.sortType == L["Groups"] then
						if string.find(bagName,L["Bank"]) then
							bagName = L["Bank"]
						elseif tonumber(bag) and bag >= 0 and bag <= 4 then
							bagName = L["Bags"]
						end
					end
					disabled = player == playerName and bagName ~= L["Mail"] and bagName:sub(1,4) ~= L["Bank"]
					if type(options.args[player].args[bagName]) ~= "table" and p.sortType ~= L["Sections"] then
						options.args[player].args[bagName] = {
							name = bagName, type = "group",
							desc = bagName,
							args = {},
							disabled = disabled,
						}
					end
					if p.sortType ~= L["Sections"] then
						opBag = options.args[player].args[bagName]
					end
					if not disabled then
						disabled = true
						for slot,item in pairs(slots) do
							disabled = false
							itemLink, count, bagid, sectionid = item[1], item[2], item[5], item[6]
							if p.sortType == L["Sections"] then
								if type(options.args[player].args[bagid]) ~= "table" then
									options.args[player].args[bagid] = {
										name = bagid, type = "group",
										desc = bagid,
										args = {},
										disabled = false,
									}
								end
								if type(options.args[player].args[bagid].args[sectionid]) ~= "table" then
									options.args[player].args[bagid].args[sectionid] = {
										name = sectionid, type = "group",
										desc = sectionid,
										args = {},
										disabled = false,
									}
								end
								opBag = options.args[player].args[bagid].args[sectionid]
							end
							if p.compressed then
								itemName = select(3,string.find(itemLink, "^|c%x+|H.+|h%[(.+)%]"))
							else
								itemName = itemLink
							end
							if type(opBag.args[itemName]) == "table" and p.compressed then
								opBag.args[itemName].count = opBag.args[itemName].count + count
								opBag.args[itemName].name = itemLink.."x"..(opBag.args[itemName].count)
								opBag.args[itemName].desc = opBag.args[itemName].desc..", {"..item[5]..", "..sectionid.."}"
							else
								opBag.args[itemName] = {
									name = itemLink.."x"..count, type = "execute",
									desc = itemLink..": {"..item[5]..", "..sectionid.."}",
									func = function() GameTooltip:SetHyperlink(itemLink) end,
									icon = item[3],
									order = 106-item[4],
									count = count,
								}
							end
						end
						if p.sortType == L["Bags"] then opBag.disabled = disabled end
					end
				else
					options.args[player].args.Money = {
						type = "execute",
						func = function() end,
						order = -1,
					}
					self:DisplayMoney(options.args[player].args.Money, slots)
				end
			end
		end --TODO: testing code, remove later
	end
end

local tableLoop = function(depth, table, func)
	local tableLoopHelper = function(depth, table, func, ...)
		if depth > 0 and type(table) == "table" then
			for k,v in pairs(table) do
				if depth == 1 then
					func(...,k,v)
				else
					tableLoopHelper(depth-1,v,func,...,k,v)
				end
			end
		end
	end
	tableLoopHelper(depth,table,func)
end

function BagginsAnywhereBags:DisplayMoney(list, money)
	local m = abacus:FormatMoneyFull(money, true)
	list.name = m
	list.desc = m
end
